public class Vector3D{
  public float x, y, z;

  public Vector3D(float x, float y, float z){
    this.x = x; 
    this.y = y; 
    this.z = z;
  }

  public Vector3D(float x, float y){
    this.x = x;
    this.y = y;
    this.z = 0f;
  }
  
  public Vector3D(){
    this.x = 0f;
    this.y = 0f;
    this.z = 0f;
  }

  public float magnitude(){
    return (float)Math.sqrt(x*x + y*y + z*z);
  }

  public void add(Vector3D v){
    x += v.x;
    y += v.y;
    z += v.z;
  }

  public void sub(Vector3D v) {
    x -= v.x;
    y -= v.y;
    z -= v.z;
  }

  public void mult(float n) {
    x *= n;
    y *= n;
    z *= n;
  }

  public void div(float n) {
    x /= n;
    y /= n;
    z /= n;
  }

  public Vector3D cross(Vector3D v) {
    float crossX = y * v.z - v.y * z;
    float crossY = z * v.x - v.z * x;
    float crossZ = x * v.y - v.x * y;
    return(new Vector3D(crossX,crossY,crossZ));
  }


  //normalize the vector
  public void normalize() {
    float m = magnitude();
    if (m > 0) {
      div(m);
    }
  }

  //limit the magnitude of the vector
  public void limit(float max) {
    if (magnitude() > max) {
      normalize();
      mult(max);
    }
  }

  //calculate 2D angle of rotation
  public float heading2D() {
    float angle = (float) Math.atan2(-y, x);
    return -1*angle;
  }

  public void setXYZ(Vector3D v) {
    x = v.x;
    y = v.y;
    z = v.z;
  }
  
  public void setXYZ(float x, float y, float z) {
    this.x = x;
    this.y = y;
    this.z = z;
  }

  public Vector3D copy() {
    return new Vector3D(x,y,z);
  }

  //static methods can be called without an instance of the class
  public static Vector3D add(Vector3D v1, Vector3D v2){
    Vector3D v = new Vector3D(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z);
    return v;
  }

  public static Vector3D sub(Vector3D v1, Vector3D v2) {
    Vector3D v = new Vector3D(v1.x - v2.x,v1.y - v2.y,v1.z - v2.z);
    return v;
  }
  public static Vector3D div(Vector3D v1, float n) {
    Vector3D v = new Vector3D(v1.x/n,v1.y/n,v1.z/n);
    return v;
  }

  public static Vector3D mult(Vector3D v1, float n) {
    Vector3D v = new Vector3D(v1.x*n,v1.y*n,v1.z*n);
    return v;
  }

  public static Vector3D copy(Vector3D v) {
    return new Vector3D(v.x, v.y,v.z);
  }

  public static float distance (Vector3D v1, Vector3D v2) {
    float dx = v1.x - v2.x;
    float dy = v1.y - v2.y;
    float dz = v1.z - v2.z;
    return (float) Math.sqrt(dx*dx + dy*dy + dz*dz);
  }
}
